#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <cryptoki.h>
#include <ctextra.h>
#include <ctutil.h>
#include <ctvdef.h>
#include <genmacro.h>

CK_CHAR g_dataToEncrypt[] = 
    { 'T', 'h', 'i', 's', ' ', 'i', 's', ' ',
      's', 'o', 'm', 'e', ' ', 'd', 'a', 't',
      'a', ' ', 't', 'o', ' ', 's', 'i', 'g',
      'n', ' ', '.', '.', '.', ' ', ' ', ' ', '\0' };

CK_CHAR g_dataToSign[] = 
    { 'T', 'h', 'i', 's', ' ', 'i', 's', ' ',
      's', 'o', 'm', 'e', ' ', 'd', 'a', 't',
      'a', ' ', 't', 'o', ' ', 's', 'i', 'g',
      'n', ' ', '.', '.', '.', ' ', ' ', '\0' };

char USERPIN[5] = { '0','0','0','0',0};

#define CHECK_CK_RV_GOTO(rv, string, label)                 \
    if (rv != CKR_OK)                                       \
    {                                                       \
        fprintf(stderr, "Error occured : %s\n", string);    \
        goto label;                                         \
    }

static CK_RV CreateSecretKeyObject(CK_SESSION_HANDLE hSession,
                                   CK_OBJECT_HANDLE_PTR phObject,
								   CK_CHAR *objLabel)
{
    CK_RV rv = CKR_OK;
    static CK_BBOOL ckTrue = TRUE;
    static CK_BBOOL ckFalse = FALSE;


    /* 
     * This is the mechanism used to generate a 3DES secret key. The fields
     * of the structure are :
     *
     *  CKM_DES3_KEY_GEN - Type of the mechanism. This informs the cryptoki 
     *                     library that we want to create a 3DES key.
     *  NULL             - This field is the parameter field. Some mechanisms 
     *                     require certain parameters to perform their 
     *                     functions. CKM_DES3_KEY_GEN does not require a 
     *                     parameter, hence the NULL value.
     *  0                - This field is the parameter length field. Since 
     *                     this mechanism type does not require a parameter,
     *                     0 is passed in as the length.
     */
	static CK_MECHANISM mechanism = {CKM_DES3_KEY_GEN, NULL, 0};
    
    /* 
     * This is the attribute template, which lays out some of the attributes the 
     * object will have when it is created. The object will also contain other
     * attributes, which are populated with default values. 
     * The attributes in the template are :
     * 
     *  CKA_LABEL - Points to a char array containing what will be the label
     *              of the key object.
     *
     *  CKA_TOKEN - Points to a CK_BBOOL variable containing the value TRUE.
     *              This object, therefore will be a token object, which 
     *              means it will persist on the token between sessions.
     *
     *  CKA_ENCRYPT - Points to a CK_BBOOL variable containing the value TRUE.
     *                This key, therefore will be able to encrypt data. 
     *
     *  CKA_DECRYPT - Points to a CK_BBOOL variable containing the value TRUE.
     *                This key, therefore will be able to decrypt data. 
     *
     *  CKA_SIGN    - Points to a CK_BBOOL variable containing the value TRUE.
     *                This key, therefore will be able to sign data. 
     *
     *  CKA_VERIFY  - Points to a CK_BBOOL variable containing the value TRUE.
     *                This key, therefore will be able to verify signatures. 
     *
     *  Other attributes are explicitly set to FALSE to ensure that they are 
     *  not set to TRUE by default.
     *
     *  The CKA_KEY_TYPE attribute is not set, since it is implied by the 
     *  mechanism. The same goes for the CKA_CLASS attribute.
     */
    CK_ATTRIBUTE objectTemplate[] = 
    {
        {CKA_LABEL,         NULL,       0},
        {CKA_TOKEN,         &ckTrue,    sizeof(CK_BBOOL)},
        {CKA_ENCRYPT,       &ckTrue,    sizeof(CK_BBOOL)},
        {CKA_DECRYPT,       &ckTrue,    sizeof(CK_BBOOL)},
        {CKA_SIGN,          &ckTrue,    sizeof(CK_BBOOL)},
        {CKA_VERIFY,        &ckTrue,    sizeof(CK_BBOOL)},
        {CKA_EXPORT,        &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_IMPORT,        &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_WRAP,          &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_UNWRAP,        &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_EXPORTABLE,    &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_EXTRACTABLE,   &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_MODIFIABLE,    &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_SENSITIVE,     &ckFalse,   sizeof(CK_BBOOL)},
        {CKA_DERIVE,        &ckFalse,   sizeof(CK_BBOOL)},
    };
    CK_SIZE objectSize = sizeof(objectTemplate) / sizeof(CK_ATTRIBUTE);

    CK_ATTRIBUTE* pAttr = NULL;

    /* Fill in the public key label */
    pAttr = FindAttribute(CKA_LABEL, objectTemplate, objectSize);
    pAttr->pValue = objLabel;
    pAttr->ulValueLen = (CK_ULONG)strlen((char*)objLabel);

    rv = C_GenerateKey(hSession, &mechanism, objectTemplate, objectSize, phObject);
    CHECK_CK_RV_GOTO(rv, "C_GenerateKey", end);

end:
    return rv;
}

static CK_RV generateKeyPair(CK_SESSION_HANDLE hSession,
                             CK_OBJECT_HANDLE* phPublicKey,
                             CK_OBJECT_HANDLE* phPrivateKey,
							 CK_CHAR *pubLabel,
							 CK_CHAR *priLabel)
{
    CK_RV rv = CKR_OK;

    CK_OBJECT_HANDLE hPublicKey = CK_INVALID_HANDLE;
    CK_OBJECT_HANDLE hPrivateKey = CK_INVALID_HANDLE;

	CK_MECHANISM mech = {CKM_RSA_PKCS_KEY_PAIR_GEN, NULL, 0};

    /*
     * Attribute values
     */
    static CK_BBOOL bTrue = TRUE;
    static CK_BBOOL bFalse = FALSE;
    static CK_ULONG keySize = 2048;

    /*
     * Attribute template for the public key.
     */
    CK_ATTRIBUTE publicKeyTpl[] = 
    {
        {CKA_TOKEN,             &bTrue,    sizeof(CK_BBOOL)},
        {CKA_PRIVATE,           &bFalse,   sizeof(CK_BBOOL)},
        {CKA_LABEL,             NULL,      0},
        {CKA_ENCRYPT,           &bTrue,    sizeof(CK_BBOOL)},
        {CKA_VERIFY,            &bTrue,    sizeof(CK_BBOOL)},
        {CKA_WRAP,              &bTrue,    sizeof(CK_BBOOL)},
        {CKA_MODIFIABLE,        &bTrue,    sizeof(CK_BBOOL)},
        {CKA_DERIVE,            &bTrue,    sizeof(CK_BBOOL)},
        {CKA_PRIME_BITS,        &keySize,  sizeof(CK_ULONG)},
    };
    CK_COUNT publicKeyTplSize = sizeof(publicKeyTpl)/sizeof(CK_ATTRIBUTE);

    /*
     * Attribute template for the private key.
     */
    CK_ATTRIBUTE privateKeyTpl[] = 
    {
        {CKA_TOKEN,             &bTrue,    sizeof(CK_BBOOL)},
        {CKA_PRIVATE,           &bFalse,   sizeof(CK_BBOOL)},
        {CKA_LABEL,             NULL,      0},
        {CKA_SENSITIVE,         &bTrue,    sizeof(CK_BBOOL)},
        {CKA_DECRYPT,           &bTrue,    sizeof(CK_BBOOL)},
        {CKA_SIGN,              &bTrue,    sizeof(CK_BBOOL)},
        {CKA_IMPORT,            &bFalse,   sizeof(CK_BBOOL)},
        {CKA_UNWRAP,            &bTrue,    sizeof(CK_BBOOL)},
        {CKA_MODIFIABLE,        &bTrue,    sizeof(CK_BBOOL)},
        {CKA_EXTRACTABLE,       &bTrue,    sizeof(CK_BBOOL)},
        {CKA_DERIVE,            &bTrue,    sizeof(CK_BBOOL)},
        {CKA_EXPORTABLE,        &bTrue,    sizeof(CK_BBOOL)},
        {CKA_SIGN_LOCAL_CERT,   &bTrue,    sizeof(CK_BBOOL)},
    };
    CK_COUNT privateKeyTplSize = sizeof(privateKeyTpl)/sizeof(CK_ATTRIBUTE);

    CK_ATTRIBUTE* pAttr = NULL;

    /* Fill in the public key label */
    pAttr = FindAttribute(CKA_LABEL, publicKeyTpl, publicKeyTplSize);
    pAttr->pValue = pubLabel;
    pAttr->ulValueLen = (CK_ULONG)strlen((char*)pubLabel);

    /* Fill in the private key label */
    pAttr = FindAttribute(CKA_LABEL, privateKeyTpl, privateKeyTplSize);
    pAttr->pValue = priLabel;
    pAttr->ulValueLen = (CK_ULONG)strlen((char*)priLabel);

    /*
     * Generate the key pair.
     */
    rv = C_GenerateKeyPair(hSession,
                           &mech, 
                           publicKeyTpl,
                           publicKeyTplSize,                            
                           privateKeyTpl,
                           privateKeyTplSize,
                           &hPublicKey,
                           &hPrivateKey);
    if (rv == CKR_OK)
    {
        *phPublicKey = hPublicKey;
        *phPrivateKey = hPrivateKey;
    }
    else
    {
        fprintf(stderr,
                "Could not generate key pair - 0x%lx (%s)",
                rv,
                strError(rv));
    }
    
    return rv;
}

static CK_RV findObject(CK_SESSION_HANDLE hSession,
                        CK_OBJECT_CLASS objClass,
                        CK_CHAR* pObjLabel,
                        CK_OBJECT_HANDLE* phObj)
{
    CK_RV rv = CKR_OK;

    /* This is the template used to search for the object. The C_FindObjects 
     * call matches all objects that have attributes matching all attributes 
     * within the search template.
     * 
     * The attributes in the search template are : 
     *  CKA_CLASS - Points to the objClass variable which contains the value
     *              CKO_SECRET_KEY, meaning this object is a secret key object.
     *  CKA_LABEL - Points to a char array containing what will be the label
     *              of the data object.
     *
     * The search will hit on all objects with the given class and label. Note
     * that it is possible to have multiple objects on a token with matching 
     * attributes, no matter what the attributes are. There is nothing 
     * precluding the existence of duplicate objects. In the case of duplicate
     * objects, the first one found is returned 
     */
    CK_ATTRIBUTE objectTemplate[] = 
    {
        {CKA_CLASS,         NULL,       0},
        {CKA_LABEL,         NULL,       0},
    };
    CK_SIZE templateSize = sizeof(objectTemplate) / sizeof(CK_ATTRIBUTE);

    CK_ULONG numObjectsToFind = 1;
    CK_ULONG numObjectsFound = 0;

    CK_ATTRIBUTE* pAttr = NULL;

    /* 
     * Fill out the template with the values to search for
     */

    /* First set the object class ... */
    pAttr = FindAttribute(CKA_CLASS, objectTemplate, templateSize);
    pAttr->pValue = &objClass;
    pAttr->ulValueLen = sizeof(CK_OBJECT_CLASS);

    /* Now set the label ... */
    pAttr = FindAttribute(CKA_LABEL, objectTemplate, templateSize);
    pAttr->pValue = pObjLabel;
    pAttr->ulValueLen = strlen((char*)pObjLabel);

    /* 
     * Now perform the search 
     */

    /* First initialise the search operation */
    rv = C_FindObjectsInit(hSession, objectTemplate, templateSize);
    CHECK_CK_RV_GOTO(rv, "C_FindObjectsInit", end);

    /* Search */
    rv = C_FindObjects(hSession,
                       phObj,
                       numObjectsToFind,
                       &numObjectsFound);
    CHECK_CK_RV_GOTO(rv, "C_FindObjects", end);

    /* Terminate the search */
    rv = C_FindObjectsFinal(hSession);
    CHECK_CK_RV_GOTO(rv, "C_FindObjects", end);

    /* Check to see if we found a matching object */
    if (numObjectsFound == 0)
    {
        fprintf(stderr, "Object not found.\n");
        rv = CKR_GENERAL_ERROR;
    }

end:
    return rv;
}

static CK_RV getMechType(CK_SESSION_HANDLE hSession,
                         CK_OBJECT_HANDLE hObj,
                         CK_MECHANISM_TYPE* pMechType)
{
    CK_RV rv = CKR_OK;

    /* Attribute template used by C_GetAttributeValue */
    static CK_KEY_TYPE keyType;
    CK_ATTRIBUTE tpl[] = 
    {
        {CKA_KEY_TYPE,      &keyType,   sizeof(keyType)}
    };

    if (hObj == CK_INVALID_HANDLE) return CKR_ARGUMENTS_BAD;

    /*
     * Determine what type of key we are dealing with.
     */
    rv = C_GetAttributeValue(hSession, hObj, tpl, 1);
    if (rv != CKR_OK) return rv;

    /*
     * We need the right encryption mechanism for the key type.
     */
    switch (keyType)
    {
        case CKK_DES:
            *pMechType = CKM_DES_ECB;
        break;

        case CKK_DES2:
        case CKK_DES3:
            *pMechType = CKM_DES3_ECB;
        break;

        case CKK_AES:
            *pMechType = CKM_AES_ECB;
        break;

        case CKK_IDEA:
            *pMechType = CKM_IDEA_ECB;
        break;
    
        case CKK_CAST:
            *pMechType = CKM_CAST_ECB;
        break;

        case CKK_RC2:
            *pMechType = CKM_RC2_ECB;
        break;

        case CKK_SEED:
            *pMechType = CKM_SEED_ECB;
        break;

        default:
            /* invalid key type */
            rv = CKR_KEY_TYPE_INCONSISTENT;
        break;
    }

    return rv;
}

CK_RV encryptData(CK_SESSION_HANDLE hSession,
                         CK_OBJECT_HANDLE hKey,
                         CK_CHAR* pData,
                         CK_SIZE dataLen,
                         CK_CHAR** ppEncData,
                         CK_SIZE* pEncDataLen)
{

    CK_RV rv = CKR_OK;

    CK_MECHANISM mech;
    CK_MECHANISM_TYPE mechType;

    /*
     * Determine and set up the encryption mechanism based on the key 
     * type of hKey
     */
    rv = getMechType(hSession, hKey, &mechType);
    if (rv != CKR_OK) return rv;

    mech.mechanism = mechType;
    mech.pParameter = NULL;
    mech.parameterLen = 0;

    /* Initialise the encrypt operation */
    rv = C_EncryptInit(hSession, &mech, hKey);
    CHECK_CK_RV_GOTO(rv, "C_EncryptInit", end);

    /* Do a length prediction so we allocate enough memory for the ciphertext */
    rv = C_Encrypt(hSession, pData, dataLen, NULL, pEncDataLen);
    CHECK_CK_RV_GOTO(rv, "C_Encrypt 1", end);
    
    *ppEncData = (CK_CHAR*)malloc(*pEncDataLen);
    if (*ppEncData == NULL) return CKR_HOST_MEMORY;

    /* Do the proper encrypt */
    rv = C_Encrypt(hSession, pData, dataLen, *ppEncData, pEncDataLen);
    CHECK_CK_RV_GOTO(rv, "C_Encrypt 2", end);

end:

    return rv;
}

static CK_RV decryptData(CK_SESSION_HANDLE hSession,
                         CK_OBJECT_HANDLE hKey,
                         CK_CHAR* pEncData,
                         CK_SIZE encDataLen,
                         CK_CHAR** ppData,
                         CK_SIZE* pDataLen)
{

    CK_RV rv = CKR_OK;

    CK_MECHANISM mech;
    CK_MECHANISM_TYPE mechType;

    /*
     * Determine and set up the decryption mechanism based on the key 
     * type of hKey
     */
    rv = getMechType(hSession, hKey, &mechType);
    if (rv != CKR_OK) return rv;

    mech.mechanism = mechType;
    mech.pParameter = NULL;
    mech.parameterLen = 0;

    /* Initialise the decrypt operation */
    rv = C_DecryptInit(hSession, &mech, hKey);
    CHECK_CK_RV_GOTO(rv, "C_DecryptInit", end);

    /* Length predication */
    rv = C_Decrypt(hSession, pEncData, encDataLen, NULL, pDataLen);
    CHECK_CK_RV_GOTO(rv, "C_Decrypt", end);

    *ppData = (CK_CHAR*)malloc(*pDataLen);
    if (*ppData == NULL) return CKR_HOST_MEMORY;

    /* Do the actual decrypt operation */
    rv = C_Decrypt(hSession, pEncData, encDataLen, *ppData, pDataLen);
    CHECK_CK_RV_GOTO(rv, "C_Decrypt", end);

end:

    return rv;
}

static CK_RV SignData(CK_SESSION_HANDLE hSession,
                      CK_OBJECT_HANDLE hKey,
                      CK_CHAR* pData,
                      CK_SIZE dataLen,
                      CK_CHAR** ppSignature,
                      CK_SIZE* pSignatureLen)
{

    CK_RV rv = CKR_OK;

    /* This is the mechanism used to sign data with RSA private key. The 
     * fields of the structure are : 
     *  CKM_RSA_PKCS  - Type of the mechanism. This informs the cryptoki 
     *                  library that we want to use the RSA_PKCS algorithm 
     *                  to perform the signature operation.
     *  NULL          - This field is the parameter field. Some mechanisms 
     *                  require certain parameters to perform their functions.
     *                  CKM_RSA_PKCS does not require a parameter, hence the 
     *                  NULL value.
     *  0             - This field is the parameter length field. Since 
     *                  this mechanism type does not require a parameter,
     *                  0 is passed in as the length.
     */
    CK_MECHANISM signMech = {CKM_RSA_PKCS, NULL, 0};

    /* Initialise the sign operation */
    rv = C_SignInit(hSession, &signMech, hKey);
    CHECK_CK_RV_GOTO(rv, "C_SignInit", end);

    /* Do a length prediction so we allocate enough memory for the signature */
    rv = C_Sign(hSession, pData, dataLen, NULL, pSignatureLen);
    CHECK_CK_RV_GOTO(rv, "C_Sign", end);
    
    *ppSignature = (CK_CHAR*)malloc(*pSignatureLen);
    if (*ppSignature == NULL) return CKR_HOST_MEMORY;

    /* Do the proper sign */
    rv = C_Sign(hSession, pData, dataLen, *ppSignature, pSignatureLen);
    CHECK_CK_RV_GOTO(rv, "C_Sign", end);

    /*
     * Note that the CKM_RSA_PKCS mechanism (as well as other RSA sign/verify
     * mechanisms) can only be part of a single-part sign operation. This
     * means that it CANNOT be called with C_SignUpdate or C_SignFinal.
     */

end:

    return rv;
}

static CK_RV VerifyData(CK_SESSION_HANDLE hSession,
                        CK_OBJECT_HANDLE hKey,
                        CK_CHAR* pData,
                        CK_SIZE dataLen,
                        CK_CHAR* pSignature,
                        CK_SIZE signatureLen)
{

    CK_RV rv = CKR_OK;

    /* This is the mechanism used to verify a signature with a public key 
     * key. The fields of the structure are : 
     *  CKM_RSA_PKCS  - Type of the mechanism. This informs the cryptoki 
     *                  library that we want to use the RSA_PKCS algorithm 
     *                  to verify the signature.
     *  NULL          - This field is the parameter field. Some mechanisms 
     *                  require certain parameters to perform their functions.
     *                  CKM_RSA_PKCS does not require a parameter, hence the 
     *                  NULL value.
     *  0             - This field is the parameter length field. Since 
     *                  this mechanism type does not require a parameter,
     *                  0 is passed in as the length.
     */
    CK_MECHANISM verifyMech = {CKM_RSA_PKCS, NULL, 0};

    /* Initialise the sign operation */
    rv = C_VerifyInit(hSession, &verifyMech, hKey);
    CHECK_CK_RV_GOTO(rv, "C_VerifyInit", end);

    /* Perform the sign operation */
    rv = C_Verify(hSession, pData, dataLen, pSignature, signatureLen);
    CHECK_CK_RV_GOTO(rv, "C_Verify", end);

    /*
     * Note that the CKM_RSA_PKCS mechanism (as well as other RSA sign/verify
     * mechanisms) it can only be part of a single-part verify operation. This
     * means that it CANNOT be called with C_VerifyUpdate or C_VerifyFinal.
     */

end:

    return rv;
}

int main() 
{
	CK_RV rv = CKR_OK;
    CK_SESSION_HANDLE hSession = CK_INVALID_HANDLE;
    CK_OBJECT_HANDLE hKey = CK_INVALID_HANDLE; 
	CK_OBJECT_HANDLE hPublicKey = CK_INVALID_HANDLE; 
    CK_OBJECT_HANDLE hPrivateKey = CK_INVALID_HANDLE;
	CK_OBJECT_HANDLE hPubKey = CK_INVALID_HANDLE; 
    CK_OBJECT_HANDLE hPriKey = CK_INVALID_HANDLE;
	CK_CHAR* pEncData = NULL;
    CK_SIZE encDataLen = 0;
    CK_CHAR* pData = NULL;
    CK_SIZE dataLen = 0;
	CK_CHAR* pSignature = NULL;
    CK_SIZE signatureLen = 0;
	CK_CHAR RandomData[16];
	int j;

    CK_SLOT_ID slotId = 0;

	printf( "Cryptoki sample application\n\n" );

    /* Initialise the cryptoki API */
	printf( "Initializing Cryptoki API ... " );
    rv = C_Initialize(NULL);
    CHECK_CK_RV_GOTO(rv, "C_Initialize", end);
	printf( "OK\n" );

	/* Check PKCS#11 version */
	printf( "Checking Cryptoki API version ... " );
	rv = CheckCryptokiVersion();
	CHECK_CK_RV_GOTO(rv, "Incompatible PKCS#11 version", end);
	printf( "OK\n" );

    /* Obtain a session so we can perform cryptoki operations */
	printf( "Obtaining a session ... " );
    rv = C_OpenSession(slotId, CKF_RW_SESSION, NULL, NULL, &hSession);
    CHECK_CK_RV_GOTO(rv, "C_OpenSession", end);
	printf( "OK\n" );

	/* Login as a user with a PIN */
	printf( "Logining as a user with a PIN ... " );
	rv = C_Login(hSession, CKU_USER, (CK_CHAR_PTR)USERPIN, (CK_SIZE) strlen((char*)USERPIN));
	CHECK_CK_RV_GOTO(rv, "C_Login as User", end);
	printf( "OK\n" );

    printf("Creating 3DES secret key ... ");
    rv = CreateSecretKeyObject(hSession, &hKey,"tdes1");
    CHECK_CK_RV_GOTO(rv, "CreateSecretKeyObject", end);
    printf("OK\n");

	printf("Creating RSA asymetric key pair ... ");
	rv = generateKeyPair(hSession, &hPublicKey, &hPrivateKey,"rsa1pub","rsa1priv");
	CHECK_CK_RV_GOTO(rv, "generateKeyPair", end);
    printf("OK\n");

	printf("Finding 3DES key ... ");
    rv = findObject(hSession, CKO_SECRET_KEY, "tdes1", &hKey);
    CHECK_CK_RV_GOTO(rv, "findObject", end);
	printf("OK\n");

    printf("Encrypting data with 3DES ... ");
    rv = encryptData(hSession, hKey, g_dataToEncrypt, (CK_SIZE)strlen((char*)g_dataToEncrypt), &pEncData, &encDataLen);
    CHECK_CK_RV_GOTO(rv, "encryptData", end);
	printf("OK\n");

    printf("Decrypting data with 3DES ... ");
    rv = decryptData(hSession, hKey, pEncData, encDataLen, &pData, &dataLen);
    CHECK_CK_RV_GOTO(rv, "decryptData", end);
	printf("OK\n");

	printf("Destrying the 3DES key ... ");
	rv = C_DestroyObject(hSession, hKey);
    CHECK_CK_RV_GOTO(rv, "DestroyObject", end);
	printf("OK\n");

	printf("Finding an RSA private key to sign ... ");
    rv = findObject(hSession, CKO_PRIVATE_KEY, "rsa1priv", &hPriKey);
    CHECK_CK_RV_GOTO(rv, "FindObject", end);
	printf("OK\n");

    printf("Performing the sign operation ... ");
    rv = SignData(hSession, hPriKey, g_dataToSign, strlen((char*)g_dataToSign), &pSignature, &signatureLen);
    CHECK_CK_RV_GOTO(rv, "SignData", end);
	printf("OK\n");

    printf("Performing the verify operation ... ");
    rv = findObject(hSession, CKO_PUBLIC_KEY, "rsa1pub", &hPubKey);
    CHECK_CK_RV_GOTO(rv, "FindObject", end);
    rv = VerifyData(hSession, hPubKey, g_dataToSign, strlen((char*)g_dataToSign), pSignature, signatureLen);
    CHECK_CK_RV_GOTO(rv, "VerifyData", end);
	printf("OK\n");

	printf("Destrying the RSA private key ... ");
	rv = C_DestroyObject(hSession, hPriKey);
    CHECK_CK_RV_GOTO(rv, "DestroyObject", end);
	printf("OK\n");

	printf("Destrying the RSA public key ... ");
	rv = C_DestroyObject(hSession, hPubKey);
    CHECK_CK_RV_GOTO(rv, "DestroyObject", end);
	printf("OK\n");

	printf("Generating random data ... ");
	C_GenerateRandom(hSession, RandomData, sizeof(RandomData));
	CHECK_CK_RV_GOTO(rv, "C_GenerateRandom", end);
	for (j=0;j<sizeof(RandomData);++j)
		printf("%02X",RandomData[j]);	
	printf("\n");

    /* We've finished our work, close the session */
	printf( "Closing the session ... " );
    rv = C_CloseSession(hSession);
    CHECK_CK_RV_GOTO(rv, "C_CloseSession", end);
	printf("OK\n");

    /* We no longer need the cryptoki API ... */
	printf( "Finilizing the cryptoki ... " );
    rv = C_Finalize(NULL);
    CHECK_CK_RV_GOTO(rv, "C_Finalize", end);
	printf("OK\n");

end:

    if (rv != CKR_OK)
    {
        fprintf(stderr,
                "Error performing create key operation : 0x%lx\n",
                rv);

        /*
         * Clean up... we don't care if there are any errors.
         */
        if (hSession != CK_INVALID_HANDLE) C_CloseSession(hSession);

        C_Finalize(NULL);
    }

    return rv;
}
